#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/ppdev.h>
#include <linux/parport.h>

#include "jtag.h"

#define CABLE_SENSE	(PARPORT_STATUS_BUSY | PARPORT_STATUS_PAPEROUT)

static int parport = -1;

int jtag_init(const char *device)
{
	int param;
	unsigned char data;

	parport = open(device, O_RDWR);
	if (parport < 0)
	{
		fprintf(stderr, "Can't open %s\n", device);
		return 0;
	}

	if (ioctl(parport, PPCLAIM) != 0)
	{
		fprintf(stderr, "Can't claim parallel port\n");
		goto fail_exit;
	}

	param = 0;
	if (ioctl(parport, PPDATADIR, &param) != 0)
	{
		fprintf(stderr, "Can't set data direction\n");
		goto fail_exit;
	}

	/* Make sure the cable is present */
	data = 0x18;
	ioctl(parport, PPWDATA, &data);

	ioctl(parport, PPRSTATUS, &data);
	if ((data & CABLE_SENSE) != PARPORT_STATUS_BUSY)
	{
		fprintf(stderr, "JTAG cable not detected (sense lines stuck high: %02x)\n", data);
		goto fail_exit;
	}

	data = 0x58;
	ioctl(parport, PPWDATA, &data);

	ioctl(parport, PPRSTATUS, &data);
	if ((data & CABLE_SENSE) != PARPORT_STATUS_PAPEROUT)
	{
		fprintf(stderr, "JTAG cable not detected (sense lines stuck low: %02x)\n", data);
		goto fail_exit;
	}

	/* Enable buffers */
	data = 0x10;
	ioctl(parport, PPWDATA, &data);

	atexit(jtag_shutdown);

	return 1;

fail_exit:
	close(parport);
	parport = -1;

	return 0;
}

void jtag_shutdown()
{
	unsigned char data;

	if (parport < 0)
		return;

	/* Disable buffers */
	data = 0x18;
	ioctl(parport, PPWDATA, &data);

	ioctl(parport, PPRELEASE);

	close(parport);
	parport = -1;
}

void jtag_tms(int value)
{
	unsigned char data;

	if (parport < 0)
		return;

	ioctl(parport, PPRDATA, &data);

	if (value)
	{
		data |= 0x04;
	} else {
		data &= ~0x04;
	}

	ioctl(parport, PPWDATA, &data);
}

void jtag_tck(int value)
{
	unsigned char data;

	if (parport < 0)
		return;

	ioctl(parport, PPRDATA, &data);

	if (value)
	{
		data |= 0x02;
	} else {
		data &= ~0x02;
	}

	ioctl(parport, PPWDATA, &data);
}

void jtag_tdi(int value)
{
	unsigned char data;

	if (parport < 0)
		return;

	ioctl(parport, PPRDATA, &data);

	if (value)
	{
		data |= 0x01;
	} else {
		data &= ~0x01;
	}

	ioctl(parport, PPWDATA, &data);
}

int jtag_tdo()
{
	unsigned char status;

	if (parport < 0)
		return 0;

	ioctl(parport, PPRSTATUS, &status);

	return (status & PARPORT_STATUS_SELECT) ? 1 : 0;
}

void jtag_clock_pulse()
{
	jtag_tck(1);
	jtag_tck(0);
}

void jtag_clock_multiple(int n)
{
	int i;

	for (i = 0; i < n; i++)
		jtag_clock_pulse();
}

