#include <stdio.h>

#include "jtag.h"

/* Sends a sequence of clock pulses with TMS values starting with LSB of seq */
void jtag_tms_sequence(int seq, int len)
{
	int i;

	for (i = 0; i < len; i++)
	{
		jtag_tms(seq & 1);
		jtag_clock_pulse();

		seq >>= 1;
	}
}

/* Moves to the TEST-LOGIC-RESET state and then to the RUN-TEST/IDLE state. */
void jtag_reset()
{
	jtag_tms_sequence(0x1f, 6);
}

/* Loads a 5-bit instruction from the RUN-TEST/IDLE state and
 * returns in the SELECT-DR-SCAN state.
 */
void jtag_instruction(unsigned char instr)
{
	int i;

	/* Move from RUN-TEST/IDLE to SHIFT-IR */
	jtag_tms_sequence(0x03, 4);

	for (i = 0; i < 5; i++)
	{
		if (i == 4)
			jtag_tms(1);
		jtag_tdi(instr & 1);
		jtag_clock_pulse();

		instr >>= 1;
	}

	/* Move from EXIT-IR to SELECT-DR-SCAN */
	jtag_tms_sequence(0x03, 2);
}

/* Starting from SELECT-DR-SCAN, reads num_bits bits.
 * 
 * If final is nonzero, exits SHIFT-DR on the last bit and returns in RUN-TEST/IDLE.
 * If final is zero, returns in SHIFT-DR so that more data can be read.
 */
void jtag_read_data(unsigned char *buf, int num_bits, int final)
{
	int i, byte, bit;

	/* Move from SELECT-DR-SCAN to SHIFT-DR */
	jtag_tms_sequence(0x00, 2);

	byte = 0;
	bit = 1;
	for (i = 0; i < num_bits; i++)
	{
		/* Clear each byte as we reach it */
		if (bit == 1)
			buf[byte] = 0;

		/* Move from SHIFT-DR to EXIT1-DR on the last bit */
		if (final && i == (num_bits - 1))
			jtag_tms(1);

		/* Read this bit */
		if (jtag_tdo())
			buf[byte] |= bit;

		/* Move to the next bit and, if necessary, the next byte */
		bit <<= 1;
		if (bit == 0x100)
		{
			bit = 1;
			byte++;
		}

		jtag_clock_pulse();
	}

	/* Move from EXIT1-DR to RUN-TEST/IDLE */
	if (final)
		jtag_tms_sequence(0x01, 2);
}

void jtag_write_u8(unsigned char data, int final)
{
	int i;

	for (i = 0; i < 8; i++)
	{
		if (final && i == 7)
			jtag_tms(1);

		jtag_tdi(data & 0x80);
		jtag_clock_pulse();

		data <<= 1;
	}

	/* Move from EXIT1-DR to RUN-TEST/IDLE */
	if (final)
		jtag_tms_sequence(0x01, 2);
}

void jtag_write_u32(unsigned int data, int final)
{
	jtag_write_u8(data >> 24, 0);
	jtag_write_u8((data >> 16) & 0xff, 0);
	jtag_write_u8((data >> 8) & 0xff, 0);
	jtag_write_u8(data & 0xff, final);
}

void bit_read_text(FILE *fp)
{
	unsigned int len, i;

	len = fgetc(fp) << 8;
	len |= fgetc(fp);

	for (i = 0; i < len; i++)
	{
		unsigned char byte = fgetc(fp);

		if (byte)
			printf("%c", byte);
	}
	printf("\n");
}

unsigned int bit_read_header(FILE *fp)
{
	unsigned int len;

	fseek(fp, 0x0d, SEEK_SET);

	while (1)
	{
		unsigned char type = fgetc(fp);

		switch (type)
		{
			case 0x61:
				printf("NCD:  ");
				bit_read_text(fp);
				break;
			case 0x62:
				printf("Part: ");
				bit_read_text(fp);
				break;
			case 0x63:
				printf("Date: ");
				bit_read_text(fp);
				break;
			case 0x64:
				printf("Time: ");
				bit_read_text(fp);
				break;
			case 0x65:
				len = fgetc(fp) << 24;
				len |= fgetc(fp) << 16;
				len |= fgetc(fp) << 8;
				len |= fgetc(fp);
				printf("%d bytes of data at 0x%x\n", len, (unsigned int)ftell(fp));
				return len;
			default:
				printf("Unknown item type 0x%02x in bit file\n", type);
				return 0;
		}
	}
}

int main(int argc, char *argv[])
{
	unsigned char buf[4];
	FILE *fp;
	unsigned int len, i;
	int ch;
	int progress, last_progress;

	if (argc != 2)
	{
		printf("Usage: xilinx_config <file.bit>\n");
		return 1;
	}

	fp = fopen(argv[1], "rb");
	if (!fp)
	{
		printf("Can't read %s\n", argv[1]);
		return 1;
	}

	len = bit_read_header(fp);
	if (!len)
	{
		fclose(fp);
		return 1;
	}
	printf("\n");

	if (!jtag_init("/dev/parport0"))
		return 1;

	jtag_reset();
	jtag_tms_sequence(0x01, 1);

	jtag_read_data(buf, 32, 0);
	if (buf[0] != 0x93 || buf[1] != 0x00 || buf[2] != 0xa2 || buf[3] != 0x90)
	{
		printf("Unsupported chip: %02x%02x%02x%02x\n", buf[3], buf[2], buf[1], buf[0]);
		return 1;
	}

	jtag_read_data(buf, 32, 1);
	if (buf[0] || buf[1] || buf[2] || buf[3])
	{
		printf("Multiple devices on the JTAG chain are not supported\n");
		return 1;
	}

	/* Shutdown sequence */
	jtag_instruction(0x05);			/* CFG_IN */
	jtag_tms_sequence(0x00, 2);
	jtag_write_u32(0x30012001, 0);
	jtag_write_u32(0x00a0bf2d, 0);
	jtag_write_u32(0x30008001, 0);
	jtag_write_u32(0x00000005, 0);
	jtag_write_u32(0x30008001, 0);
	jtag_write_u32(0x00000007, 0);
	jtag_write_u32(0x00000000, 1);

	jtag_instruction(0x0c);			/* JSTART */
	jtag_tms_sequence(0x00, 2);
	jtag_clock_multiple(12);
	jtag_tms_sequence(0x03, 3);

	jtag_instruction(0x05);			/* CFG_IN */
	jtag_tms_sequence(0x00, 2);
	jtag_write_u32(0x30008001, 0);
	jtag_write_u32(0x00000008, 0);
	jtag_write_u32(0x30012001, 0);
	jtag_write_u32(0x00a03f2d, 0);
	jtag_write_u32(0x00000000, 1);

	/* CFG_IN */
	jtag_instruction(0x05);
	jtag_tms_sequence(0x00, 2);

	printf("|");
	for (i = 0; i < 50; i++)
		printf("-");
	printf("|");
	last_progress = -1;

	for (i = 0; i < len; i++)
	{
		progress = i * 100 / (len - 1);
		if (progress != last_progress)
		{
			int j;

			printf("\r|");
			for (j = 0; j < (progress / 2); j++)
				printf("-");
			printf("*");
			for (; j < 50; j++)
				printf("-");
			printf("| %d%%", progress);
			fflush(stdout);

			last_progress = progress;
		}

		ch = fgetc(fp);
		if (ch == EOF)
		{
			printf("\n\n*** Truncated bitstream\n");
			return 1;
		}

		jtag_write_u8(ch, i == (len - 1));
	}
	printf("\n\n");

	/* JSTART */
	jtag_instruction(0x0c);
	jtag_tms_sequence(0x00, 2);
	jtag_clock_multiple(12);
	jtag_tms_sequence(0x03, 3);

	/* Get status */
	jtag_instruction(0x05);
	jtag_tms_sequence(0x00, 2);
	jtag_write_u32(0x2800e001, 0);
	jtag_write_u32(0x00000000, 1);

	jtag_instruction(0x04);
	jtag_read_data(buf, 32, 1);
	if (buf[2] & 0x04)
		printf("Programmed successfully\n");
	else printf("*** Configuration error?\n");

	/* Get usercode */
	jtag_instruction(0x08);
	jtag_read_data(buf, 32, 1);
	printf("Usercode %02x%02x%02x%02x\n", buf[3], buf[2], buf[1], buf[0]);

	return 0;
}

